;*******************************************************************
;                                                                  *
;    Filename:	    koji.asm                                     *
;    Date:          2009.06.29.                                    *
;    File Version:  1.1                                            *
;                                                                  *
;    Author:        M.Urushidani                                   *
;    Update:		7/1,7/3,7/4,7/5,7/6,7/9,7/13                                                          *
;*******************************************************************
;    Files required:   p16f873A.inc                                *
;*******************************************************************
;    Notes:     Use Assembler 'MPASM'                              *
;*******************************************************************

	list		p=16f873A		; list directive to define processor
	errorlevel	-302
	; processor specific variable definitions
	#include	<p16f873A.inc>	
	
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _HS_OSC & _WRT_OFF & _LVP_OFF

;******************* VARIABLE DEFINITIONS **************************
TempC		EQU	0x20		;temp general purpose regs
TempD		EQU	0x21
TempE		EQU	0x22
Cnt_100mS	EQU	0x23		;100mS counter
MsdTime		EQU	0x26		;most significant Timer
LsdTime		EQU	0x27		;Least significant Timer
status_temp	EQU	0x28		;interrupt push
PABuf		EQU	0x29
PBBuf		EQU	0x2A
ADFlag		EQU	0x2B
notuse1		EQU	0x2C
notuse2		EQU	0x2D
notuse3		EQU	0x2E
MsdTemp		EQU	0x2F
LsdTemp		EQU	0x30
Cnt_mS		EQU	0x31		;2mS counter
Cnt_Sec		EQU 0x32		;1S counter
SysFlag		EQU	0x33
counter		EQU 0x34
count		EQU	0x35
temp		EQU	0x36
H_byte		EQU	0x37
L_byte		EQU	0x38
R0			EQU	0x39
R1			EQU	0x3A
R2			EQU	0x3B
ADval		EQU	0x40
tbl_cnt		EQU	0x41
T_val1		EQU	0x42
T_val2		EQU	0x43
_AL			EQU	0x44
_AH			EQU	0x45
_BL			EQU	0x46
_BH			EQU 0x47
ADfrac		EQU	0x48
WORK1		EQU	0x49
WORK2		EQU	0x4A
Tgt_L		EQU	0x4B
Tgt_H		EQU	0x4C
LapS_L		EQU	0x4D
LapS_H		EQU	0x4E
Tmp_H		EQU	0x4F
Tmp_L		EQU	0x50
ST_TEMP		EQU	0x51
PL_TEMP		EQU	0x52
FS_TEMP		EQU	0x53
W_TEMP		EQU	0x54
Dif_H		EQU	0x55
Dif_L		EQU	0x56
;**** limit 0x7f ****
OptionReg	EQU	1
PCL			EQU	2
BcdMsd		EQU	26
Bcd			EQU	27
ADOver		EQU	5			;bit 5 --> a/d over
TempUp		EQU	7			;SysFlag.TempUp=1 then heater on, else off.
;
push		macro
			movwf	W_TEMP
			swapf	STATUS,W
			clrf	STATUS
			movwf	ST_TEMP
			movf	FSR,W
			movwf	FS_TEMP
			endm

pop			macro
			movf	FS_TEMP,W
			movwf	FSR
			swapf	ST_TEMP,W
			movwf	STATUS
			swapf	W_TEMP,F
			swapf	W_TEMP,W
			endm						

;
;*******************************************************************
	ORG     0x000           	; processor reset vector
	GOTO	start				;skip over interrupt vector
;*******************************************************************
	ORG	0x004			;interrupt vector
	push
	call	ServiceInterrupts
	pop
	retfie
;************************* ROM TABLE *******************************
tbl_therm				;2,3,4,...,49,50deg
	addwf	PCL,f
	dt	.254,.248,.241,.235,.229,.222,.216,.210,.204,.198,.193,.187
	dt	.181,.176,.171,.165,.160,.155,.151,.146,.141,.137,.132,.128
	dt	.124,.120,.116,.112,.108,.105,.101,.98,.95,.92,.88,.85,.83
	dt	.80,.77,.75,.72,.70,.67,.65,.63,.61,.59,.57,.55 

;************************* interrupt routine ***********************
ServiceInterrupts
	btfsc	INTCON,INTF			;RB0-INT?
	goto	ServiceRBINT
	btfsc	INTCON,T0IF			;rtcc(TMR0) interrupt?
	goto	ServiceTMR0			;yes, then service
	btfsc	PIR1,TMR1IF			;TMR1 interrupt?
	goto	ServiceTMR1			;yes, then service
	return

ServiceRBINT
	bcf		INTCON,INTF
	bsf		PORTA,5				;Heater off
	movlw	0xE0				;45536=0xB1E0,preload
	movwf	TMR1L
	movlw	0xB1
	movwf	TMR1H
	bsf		STATUS,RP0			;bank1
	bsf		PIE1,TMR1IE			;enable timer1 int
	bcf		STATUS,RP0			;bank0
	bsf		T1CON,TMR1ON		;timer1 on
	bsf		INTCON,PEIE
	return
	
ServiceTMR0						;every 5mS
	movlw	.178				;initialize rtcc
	movwf	TMR0
	bcf		INTCON,T0IF			;clr int flag
	call	IncTimer			;inc timer
	btfsc	PORTC,0				;last digit?
	call	SampleAd			;then sample a/d
	call	UpdateDisplay		;update display
	bsf		INTCON,T0IE
	return

ServiceTMR1
	bcf		PIR1,TMR1IF			;clr int flag
	btfsc	SysFlag,TempUp		;temperature up?
	bcf		PORTA,5				;heater on
	return
	
SampleAd
	call	DoAd				;do a ad conversion
AdDone
	btfsc	ADCON0,GO			;ad done?
	goto	AdDone				;no, then loop
	bsf		ADFlag,ADOver		;set a/d over flag
	return

DoAd
	bsf		ADCON0,ADON			;start a/d
	call	time15				;wait acquisition time
	bsf		ADCON0,GO			;start conversion
	return

;The display is incremented every 200*5mS=1 Sec.
IncTimer
	incf	Cnt_mS,W			;inc 2mS counter for check
	xorlw	.50					;=50?, 50*2=100mS?
	btfsc	STATUS,Z			;skip if clr(not eq)
	goto	DoUpCnt				;Z=1, then inc 2mS counter
	incf	Cnt_mS,F			;inc 2mS counter really
	return
	
DoUpCnt
	clrf	Cnt_mS				;clr 2mS counter
	incf	Cnt_100mS,W			;inc 100mS counter
	xorlw	.10					;=10?, 10*100=1000mS=1S?
	btfsc	STATUS,Z			;skip if clr(not eq)
	goto	DoIncSec			;Z=1, then inc Sec
	incf	Cnt_100mS,F			;inc 100mS counter really
	return
	
DoIncSec
	incf	LapS_L,F			;++lap sec counter
	btfsc	STATUS,Z			;carry?
	incf	LapS_H,F			;yes
	movf	LapS_H,W
	xorlw	0x03				;360h=864?
	btfss	STATUS,Z
	goto	_next0				;no
	movf	LapS_L,W
	xorlw	0x60
	btfss	STATUS,Z
	goto	_next0
	clrf	LapS_H
	clrf	LapS_L
	decf	Tgt_L
	btfsc	STATUS,Z			;borrow?
	decf	Tgt_H				;yes
	movf	Tgt_L,W
	sublw	0x2C	;0x5E		;W=<30.0deg?	35.0deg?
	btfss	STATUS,C
	goto	_next0				;W>0x5E
	movlw	0x2C	;0x5E		;0x5E>=Tgt_L
	movwf	Tgt_L				;then Tgt_L=0x5E
	
_next0	
	btfsc	SysFlag,1
	goto	_next1
	bsf		SysFlag,1
	goto	_next2
_next1
	bcf		SysFlag,1
_next2
	clrf	Cnt_100mS			;clr 100mS counter
	incf	Cnt_Sec,W			;inc 1S counter
	xorlw	.60					;=60?, 60S?
	btfsc	STATUS,Z			;skip if clr(not eq)
	goto	DoIncTime			;Z=1, then inc Min
	incf	Cnt_Sec,F			;inc 1S counter really
	return
	
DoIncTime
	clrf	Cnt_Sec				;clr count
	incf	LsdTime,W			;get lsd
	andlw	0x0F				;mask high nibble
	xorlw	0x0a				;=10?
	btfsc	STATUS,Z			;no, then skip
	goto	IncSecondLsd		;inc next lsd
	incf	LsdTime,F			;else, inc timer
	return
	
IncSecondLsd
	swapf	LsdTime,W			;get hi in low nibble
	andlw	0x0F				;mask high nibble
	addlw	1					;inc it
	movwf	LsdTime				;restore back
	swapf	LsdTime,F
	xorlw	0x06				;60min?
	btfsc	STATUS,Z			;no, then skip
	goto	IncThirdLsd			;else, inc next lsd
	return
	
IncThirdLsd
	clrf	LsdTime
	incf	MsdTime,W			;get 3rd lsd
	andlw	0x0f				;mask high nibble
	xorlw	0x0a				;=10?
	btfsc	STATUS,Z			;no, then skip
	goto	IncMsd				;else, Msd
	incf	MsdTime,F			;else, inc Timer
	return
	
IncMsd
	swapf	MsdTime,W			;get hi in lo nibble
	andlw	0x0F				;mask high nible
	addlw	1					;inc timer
	movwf	MsdTime				;restore back
	swapf	MsdTime,F
	xorlw	0x0a				;=10?
	btfsc	STATUS,Z			;no, then skip
	clrf	MsdTime				;clr msd
	return
	
UpdateDisplay	;***************************************************
	bcf		SysFlag,0			;dot off
	movf	PORTC,W				;present sink value in W
	andlw	B'01111111'			;B'00001111'
	movwf	TempC				;save sink value in tempC	
	movlw	B'01111111'			;disable all digits sinks
	movwf	PORTC
	bcf		TempC,7				;preset for lsd sink
	rrf		TempC,F				;determine next sink value
	btfsc	STATUS,C			;C=0?
	bsf		TempC,6				;no, then reset LSD sink
	btfss	TempC,0				;else, see if Msd
	goto	UpdateMsd			;yes, then do Msd
	btfss	TempC,1				;see if 3rd LSD
	goto	Update3rdLsd		;yes, then do 3rd Lsd
	btfss	TempC,2				;see if 2nd Lsd
	goto	Update2ndLsd		;yes, then do 2nd lsd
	btfss	TempC,3
	goto	UpdateLsd
	btfss	TempC,4
	goto	UpdateMsd_T
	btfss	TempC,5
	goto	UpdateMid_T

UpdateLsd_T
	movf	LsdTemp,W
	andlw	0x0f
	goto	DisplayOut
	
UpdateMid_T
	bsf		SysFlag,0			;dot on
	swapf	LsdTemp,W
	andlw	0x0f
	goto	DisplayOut
	
UpdateMsd_T
	movf	MsdTemp,W	
	andlw	0x0f
	goto	DisplayOut	
	
UpdateLsd
	movf	LsdTime,W			;get Lsd in W
	andlw	0x0f
	goto	DisplayOut			;enable display
	
Update2ndLsd
	swapf	LsdTime,W			;get 2nd Lsd in W
	andlw	0x0F				;mask rest
	goto	DisplayOut			;enable display
	
Update3rdLsd
	btfsc	SysFlag,1
	bsf		SysFlag,0			;dot on
	movf	MsdTime,W			;get 3rd Lsd in W
	andlw	0x0F				;mask low nibble
	goto	DisplayOut			;enable display
	
UpdateMsd
	swapf	MsdTime,W			;get Msd in W
	andlw	0x0F				;mask rest
	
DisplayOut
	Call	LedTable			;get digit output
	movwf	PORTB				;drive leds
	bcf		TempC,7
	btfsc	SysFlag,0
	bsf		TempC,7
	movf	TempC,W
	movwf	PORTC

	return
	
LedTable
	addwf	PCL,F				;add to PC low
	retlw	B'01111110'			;0	gfedcbaz
	retlw	B'00001100'			;1
	retlw	B'10110110'			;2
	retlw	B'10011110'			;3
	retlw	B'11001100'			;4
	retlw	B'11011010'			;5
	retlw	B'11111010'			;6
	retlw	B'00001110'			;7
	retlw	B'11111110'			;8
	retlw	B'11001110'			;9
	retlw	B'00000000'			;blank
	
Chk2LsdZero
	call	ChkMsdZero			;msd=0?
	btfss	STATUS,Z			;yes, then skip
	return						;else, return
	swapf	LsdTime,W			;get 2nd lsd
	andlw	0x0f				;mask of LSD
	btfss	STATUS,Z			;0?, then skip
	return
	retlw	.10					;else, return with 10
	
ChkMsdZero
	movf	MsdTime,W			;get Msd in W
	btfss	STATUS,Z			;=0?, then skip
	return
	retlw	.10					;ret with 10

;********************* subroutine in interrupt *********************
time15
	MOVLW	1Ah					;wait15usec
	MOVWF	counter
t_lp1
	DECFSZ	counter,F			;2+3*25-1=76
	GOTO	t_lp1
	RETURN						;=76*.2us		=38*0.4us

;************************** reset start ****************************

start
	call	InitPorts
	call	InitTimers
	call	InitAd
	movlw	0x01				;target temp.=40.0deg
	movwf	Tgt_H				;0x190=400
	movlw	0x90
	movwf	Tgt_L
	clrf	LapS_H				;lap timer reset
	clrf	LapS_L
loop
	btfsc	ADFlag,ADOver		;a/d over?
	call	UpdateAd			;yes then update
	goto	loop
	
UpdateAd
	btfss	PIR1,ADIF			;a/d done?
	return						;no, then leave
	clrf	H_byte
	clrf	L_byte
	BSF		STATUS,RP0			;bank1
	movf	ADRESL,W			;get a/d value
	BCF		STATUS,RP0			;bank0
	movwf	ADval
	;limit check
	xorlw	.255
	btfsc	STATUS,Z
	goto	_upper_lim
	goto	_chk_lower
_upper_lim
	movlw	.18
	movwf	L_byte
	goto	_set_val
_chk_lower
	movf	ADval,W
	sublw	.55					;ADval=<55?
	btfsc	STATUS,C
	goto	_lower_lim			;55>=ADval
	goto	_chk_table			;55<ADval
_lower_lim
	movlw	0xf4				;500=0x01f4...50.0deg
	movwf	L_byte
	movlw	0x01
	movwf	H_byte
	goto	_set_val
_chk_table
	clrf	tbl_cnt
_chk_loop
	movf	tbl_cnt,W
	call	tbl_therm
	movwf	T_val1
	subwf	ADval,W				;ADval-W->W
	movwf	ADfrac				;ADval-T_val1
	btfss	STATUS,C
	goto	_next_chk			;ADval<W
	goto	_temp_set			;ADval>=W
_next_chk
	incf	tbl_cnt,F
	movf	tbl_cnt,W
	xorlw	.48
	btfsc	STATUS,Z
	goto	_lower_lim			;tbl_cnt=48
	goto	_chk_loop
_temp_set
	movf	tbl_cnt,W
	addlw	2					;W=temperature
	movwf	_AL
	clrf	_AH
	call	mult_10				;temp*10
	movf	_AL,W
	movwf	L_byte
	movf	_AH,W
	movwf	H_byte

;fraction adjust
	movf	ADfrac,W			;ADval-Tval1
	btfsc	STATUS,Z
	goto	_set_val			;ADfrac=0, then no adj.
	decf	tbl_cnt,W
	call	tbl_therm
	movwf	T_val2
	movf	T_val1,W
	subwf	T_val2,W			;Tval2-Tval1->W
	sublw	7					;7-W->W
	btfsc	STATUS,Z			;0?
	addlw	1					;=0, then +1
	movwf	WORK1				;fraction
	movf	ADval,W
	subwf	T_val2,W			;T_val2-ADval
	movwf	ADfrac
	clrf	WORK2
_loop_frac
	movf	WORK1,W
	addwf	WORK2,F
	decfsz	ADfrac,F
	goto	_loop_frac
	movlw	.10
	subwf	L_byte,F			;L_byte-10->L_byte
	btfss	STATUS,C			;skip if C=1(no borrow)
	decf	H_byte				;C=0(borrow)
	movf	WORK2,W
	addwf	L_byte,F
	btfsc	STATUS,C
	incf	H_byte				;C=1(carry)
	
	movf	H_byte,W			;Copy to Tmp
	movwf	Tmp_H
	movf	L_byte,W
	movwf	Tmp_L
	
	bcf		SysFlag,TempUp		;heater off
	movf	Tmp_L,W				;compare
	subwf	Tgt_L,W				;W=Tgt_L-Tmp_L
	movwf	Dif_L				;Dif_L=W
	movf	Tmp_H,W
	btfss	STATUS,C
	incf	Tmp_H,W				;W=Tgt_H-(Tmp_H(+1))
	subwf	Tgt_H,W				;W=Tgt_H-W
	movwf	Dif_H				;Dif_H=W
	btfss	Dif_H,7				;sign check
	bsf		SysFlag,TempUp		;plus,heater on

;**************************************************		
_set_val						;set H_byte,L_byte
	call	B2_BCD
	movf	R2,W				;get Lsd
	movwf	LsdTemp				;save in LSD
	movf	R1,W				;get Msd
	movwf	MsdTemp				;save in Msd
	bcf		PIR1,ADIF			;clr interrupt flag
	bcf		ADCON0,ADON			;turn off a/d
	bcf		ADFlag,ADOver		;clr flag
	return

mult_10							;A*10->A
	bcf		STATUS,C			;A*2
	rlf		_AL,F
	rlf		_AH,F
	movf	_AL,W				;A*2->B
	movwf	_BL
	movf	_AH,W
	movwf	_BH
	bcf		STATUS,C			;A*4
	rlf		_AL,F
	rlf		_AH,F
	bcf		STATUS,C			;A*8
	rlf		_AL,F
	rlf		_AH,F
	movf	_BL,W				;B+A
	addwf	_AL,F
	movf	_BH,W
	btfsc	STATUS,C
	incf	_BH,W
	addwf	_AH,F
	return
					

InitPorts
;*** Port setting ***
	BSF		STATUS,RP0		;bank1
	movlw	B'00000001'		;make RB0:input, others outputs
	movwf	TRISB
	clrf	TRISC			;make RC0-7 outputs
	movlw	B'10001110'		;make AN0=A/D,others=Digital
	movwf	ADCON1
	movlw	B'00000001'		;make RA1-5 outputs, RA0=input(A/D)
	movwf	TRISA
	bcf		STATUS,RP0		;bank0
	movlw	B'00100000'		;triac=off
	movwf	PORTA
	clrf	PORTB			;allSEG=off
	movlw	B'01111111'		;timeLED=off, thermLED=off, dot=off
	movwf	PORTC
	bcf		PORTC,6			;enable time-MSB digit power
	return
	
;The clock speed is 20MHz. Dividing internal clk. by a 128
;prescaler, the rtcc(timer0) will be incremented every 25.6us.
;If rtcc is preloaded with 60. it will take (256-60)*25.6us
;to overflow i.e. 5mS. So the end result is that we get a rtcc
;interrupt every 5mS.

InitTimers
							;*** Timer1 setting ***
	movlw	B'00000100'		;prescale1:1,TMR1:stop
	movwf	T1CON
	bsf		INTCON,PEIE
							;*** Timer0 setting ***
	clrf	MsdTime			;clr timers
	clrf	LsdTime
	bsf		STATUS,RP0		;select pg 1
	movlw	B'11000110'		;assign ps to rtcc(TMR0)
	movwf	OptionReg		;ps=128
	bcf		STATUS,RP0		;select pg 0
	movlw	B'00110000'		;enalbe rtcc(TMR0),INT interrupt
	movwf	INTCON
	movlw	.178			;preload rtcc(TMR0)
	movwf	TMR0			;start counter
	retfie
	
InitAd
	bcf		STATUS,RP0		;bank0
	movlw	0x80			;f/32,ch0,ADoff
	movwf	ADCON0
	return

B2_BCD
	bcf		STATUS,0		;clear the carry bit
	movlw	.16
	movwf	count
	clrf	R0
	clrf	R1
	clrf	R2
loop16
	rlf		L_byte,F
	rlf		H_byte,F
	rlf		R2,F
	rlf		R1,F
	rlf		R0,F
	
	decfsz	count,F
	goto	adjDEC
	retlw	0
	
adjDEC
	movlw	R2
	movwf	FSR
	call	adjBCD
	
	movlw	R1
	movwf	FSR
	call	adjBCD
	
	movlw	R0
	movwf	FSR
	call	adjBCD
	
	goto	loop16
	
adjBCD
	movlw	3
	addwf	0,W
	movwf	temp
	btfsc	temp,3			;test if result>7
	movwf	0
	movlw	30
	addwf	0,W
	movwf	temp
	btfsc	temp,7			;test if result>7
	movwf	0				;save as MSD
	retlw	0
	
	END                     ; directive 'end of program'

